/* 
   check-usb-scanner -- Find out what USB scanner chipset is used

   Copyright (C) 2003 Henning Meier-Geinitz <henning@meier-geinitz.de>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
*/

#include <stdio.h>
#include <usb.h>
#include <string.h>
#include <errno.h>

#define BUILD 4
/* #define DEBUG */

/* Get USB string descriptor and convert to ASCII (dirty hack) */
char *
get_string_descriptor (struct usb_device *dev, int index)
{
  usb_dev_handle *handle;
  char *buffer;
  struct usb_string_descriptor *sd;
  int size = 100;
  int i;

  if (!index)
    return 0;
  buffer = calloc (1, size + 1);
  if (!buffer)
    return 0;
  sd = (struct usb_string_descriptor *) buffer;
  handle = usb_open (dev);
  if (!handle)
    return 0;
  if (usb_control_msg (handle,
		       USB_ENDPOINT_IN + USB_TYPE_STANDARD + USB_RECIP_DEVICE,
		       USB_REQ_GET_DESCRIPTOR,
		       (USB_DT_STRING << 8) + index, 0, buffer,
		       size, 100) < 0)
    return 0;
  if (sd->bLength < 2 || sd->bLength > size
      || sd->bDescriptorType != USB_DT_STRING)
    return 0;
  size = sd->bLength - 2;
  for (i = 0; i < (size / 2); i++)
    buffer[i] = buffer[2 + 2 * i];
  buffer[i] = 0;

  return buffer;
}

/* Send a control message to a GT-68XX device */
int
gt68xx_send_control_msg (struct usb_device *dev, int is_gt6816)
{
  char req[64];
  usb_dev_handle *handle;
  int result;

  handle = usb_open (dev);
  if (handle == 0)
    {
      printf ("  Couldn't open device. Unload the scanner driver and run as root.\n");
      return 0;
    }

  result = usb_claim_interface (handle, 0);
  if (result < 0)
    {
      printf ("  Couldn't claim interface. Unload the scanner driver and run as root.\n");
      usb_close (handle);
      return 0;
    }

  memset (req, 0, 64);
  req[0] = 0x2e; /* get identification information */
  req[1] = 0x01;

  result = usb_control_msg (handle, 0x40, is_gt6816 ? 0x04 : 0x01, 0x2010, 0x3f40,
			    req, 64, 1000);
  if (result <= 0)
    {
      printf ("    Couldn't send write control message (%s)\n", strerror (errno));
      usb_close (handle);
      return 0;
    }
  result = usb_control_msg (handle, 0xc0, is_gt6816 ? 0x04 : 0x01, 0x2011, 0x3f00,
			    req, 64, 1000);
  if (result <= 0)
    {
      printf ("    Couldn't send read control message (%s)\n", strerror (errno));
      usb_close (handle);
      return 0;
    }

  if (req[0] != 0 || (req[1] != 0x2e && req[1] != 0))
    {
      printf ("    Unexpected result from control message (%0x/%0x)\n", req[0], req[1]);
      usb_close (handle);
      return 0;
    }
  usb_close (handle);
  return 1;
}

/* Check for Grandtech GT-6801 */
int
check_gt6801 (struct usb_device *dev)
{
  printf ("    checking for GT-6801 ...\n");

  /* Check device descriptor */
  if (dev->descriptor.bDeviceClass != USB_CLASS_VENDOR_SPEC)
    {
      printf ("    this is not a GT-6801 (bDeviceClass = %d)\n",
	      dev->descriptor.bDeviceClass);
      return 0;
    }
  if (dev->descriptor.bcdUSB != 0x110)
    {
      printf ("    this is not a GT-6801 (bcdUSB = 0x%x)\n",
	      dev->descriptor.bcdUSB);
      return 0;
    }
  if (dev->descriptor.bDeviceSubClass != 0xff)
    {
      printf ("    this is not a GT-6801 (bDeviceSubClass = 0x%x)\n",
	      dev->descriptor.bDeviceSubClass);
      return 0;
    }
  if (dev->descriptor.bDeviceProtocol != 0xff)
    {
      printf ("    this is not a GT-6801 (bDeviceProtocol = 0x%x)\n",
	      dev->descriptor.bDeviceProtocol);
      return 0;
    }

  /* Check endpoints */
  if (dev->config[0].interface[0].altsetting[0].bNumEndpoints
      != 1)
    {
      printf ("    this is not a GT-6801 (bNumEndpoints = %d)\n",
	      dev->config[0].interface[0].altsetting[0].bNumEndpoints);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress != 0x81)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize != 0x40)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval != 0x00))
    {
      printf ("    this is not a GT-6801 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval);
      return 0;
    }
  if (!gt68xx_send_control_msg (dev, 0))
    return 0;
  printf ("  FOUND: Chipset is Grandtech GT-6801\n");
  return 1;
}

/* Check for Grandtech GT-6816 */
int
check_gt6816 (struct usb_device *dev)
{
  printf ("    checking for GT-6816 ...\n");

  /* Check device descriptor */
  if ((dev->descriptor.bDeviceClass != USB_CLASS_PER_INTERFACE)
      || (dev->config[0].interface[0].altsetting[0].bInterfaceClass !=
	  USB_CLASS_VENDOR_SPEC))
    {
      printf ("    this is not a GT-6816 (bDeviceClass = %d, bInterfaceClass = %d)\n",
	      dev->descriptor.bDeviceClass,
	      dev->config[0].interface[0].altsetting[0].bInterfaceClass);
      return 0;
    }
  if (dev->descriptor.bcdUSB != 0x110)
    {
      printf ("    this is not a GT-6816 (bcdUSB = 0x%x)\n",
	      dev->descriptor.bcdUSB);
      return 0;
    }
  if (dev->descriptor.bDeviceSubClass != 0x00)
    {
      printf ("    this is not a GT-6816 (bDeviceSubClass = 0x%x)\n",
	      dev->descriptor.bDeviceSubClass);
      return 0;
    }
  if (dev->descriptor.bDeviceProtocol != 0x00)
    {
      printf ("    this is not a GT-6816 (bDeviceProtocol = 0x%x)\n",
	      dev->descriptor.bDeviceProtocol);
      return 0;
    }

  /* Check endpoints */
  if (dev->config[0].interface[0].altsetting[0].bNumEndpoints
      != 2)
    {
      printf ("    this is not a GT-6816 (bNumEndpoints = %d)\n",
	      dev->config[0].interface[0].altsetting[0].bNumEndpoints);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress != 0x81)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize != 0x40)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval != 0x00))
    {
      printf ("    this is not a GT-6816 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[1].bEndpointAddress != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].wMaxPacketSize != 0x40)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].bInterval != 0x00))
    {
      printf ("    this is not a GT-6816 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bInterval);
      return 0;
      
    }
  if (!gt68xx_send_control_msg (dev, 1))
    return 0;
  printf ("  FOUND: Chipset is Grandtech GT-6816\n");
  return 1;
}

/* Check for Mustek MA-1017 */
int
check_ma1017 (struct usb_device *dev)
{
  char req[2];
  usb_dev_handle *handle;
  int result;
  unsigned char res;

  printf ("    checking for MA-1017 ...\n");

  /* Check device descriptor */
  if ((dev->descriptor.bDeviceClass != USB_CLASS_PER_INTERFACE)
      || (dev->config[0].interface[0].altsetting[0].bInterfaceClass !=
	  USB_CLASS_PER_INTERFACE))
    {
      printf ("    this is not a MA-1017 (bDeviceClass = %d, bInterfaceClass = %d)\n",
	      dev->descriptor.bDeviceClass,
	      dev->config[0].interface[0].altsetting[0].bInterfaceClass);
      return 0;
    }
  if (dev->descriptor.bcdUSB != 0x100)
    {
      printf ("    this is not a MA-1017 (bcdUSB = 0x%x)\n",
	      dev->descriptor.bcdUSB);
      return 0;
    }
  if (dev->descriptor.bDeviceSubClass != 0x00)
    {
      printf ("    this is not a MA-1017 (bDeviceSubClass = 0x%x)\n",
	      dev->descriptor.bDeviceSubClass);
      return 0;
    }
  if (dev->descriptor.bDeviceProtocol != 0x00)
    {
      printf ("    this is not a MA-1017 (bDeviceProtocol = 0x%x)\n",
	      dev->descriptor.bDeviceProtocol);
      return 0;
    }

  /* Check endpoints */
  if (dev->config[0].interface[0].altsetting[0].bNumEndpoints
      != 3)
    {
      printf ("    this is not a MA-1017 (bNumEndpoints = %d)\n",
	      dev->config[0].interface[0].altsetting[0].bNumEndpoints);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress != 0x01)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval != 0x00))
    {
      printf ("    this is not a MA-1017 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[1].bEndpointAddress != 0x82)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].wMaxPacketSize != 0x40)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].bInterval != 0x00))
    {
      printf ("    this is not a MA-1017 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bInterval);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[2].bEndpointAddress != 0x83)
      || (dev->config[0].interface[0].altsetting[0].endpoint[2].bmAttributes != 0x03)
      || (dev->config[0].interface[0].altsetting[0].endpoint[2].wMaxPacketSize != 0x01)
      || (dev->config[0].interface[0].altsetting[0].endpoint[2].bInterval != 0x01))
    {
      printf ("    this is not a MA-1017 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[2].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[2].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[2].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[2].bInterval);
      return 0;
    }

  /* read a register value */
  handle = usb_open (dev);
  if (handle == 0)
    {
      printf ("    Couldn't open device. Unload the scanner driver and run as root.\n");
      return 0;
    }
  result = usb_claim_interface (handle, 0);
  if (result < 0)
    {
      printf ("    Couldn't claim interface. Unload the scanner driver and run as root.\n");
      usb_close (handle);
      return 0;
    }

  req[0] = 0x00;
  req[1] = 0x02 | 0x20;
  result = usb_bulk_write (handle, 0x01, req, 2, 1000);
  if (result <= 0)
    {
      printf ("    this is not a MA-1017 (Error during bulk write)\n");
      usb_close (handle);
      return 0;
    }
  result = usb_bulk_read (handle, 0x82, &res, 1, 1000);
  if (result <= 0)
    {
      printf ("    this is not a MA-1017 (Error during bulk read)\n");
      usb_close (handle);
      return 0;
    }
  usb_close (handle);
  printf ("  FOUND: Chipset is Mustek MA-1017\n");
  return 1;
}

/* Check for Mustek MA-1015 */
int
check_ma1015 (struct usb_device *dev)
{
  char req[8];
  usb_dev_handle *handle;
  int result;
  unsigned char res;

  printf ("    checking for MA-1015 ...\n");

  /* Check device descriptor */
  if (dev->descriptor.bDeviceClass != USB_CLASS_VENDOR_SPEC)
    {
      printf ("    this is not a MA-1015 (bDeviceClass = %d)\n",
	      dev->descriptor.bDeviceClass);
      return 0;
    }
  if (dev->descriptor.bcdUSB != 0x100)
    {
      printf ("    this is not a MA-1015 (bcdUSB = 0x%x)\n",
	      dev->descriptor.bcdUSB);
      return 0;
    }
  if (dev->descriptor.bDeviceSubClass != 0xff)
    {
      printf ("    this is not a MA-1015 (bDeviceSubClass = 0x%x)\n",
	      dev->descriptor.bDeviceSubClass);
      return 0;
    }
  if (dev->descriptor.bDeviceProtocol != 0xff)
    {
      printf ("    this is not a MA-1015 (bDeviceProtocol = 0x%x)\n",
	      dev->descriptor.bDeviceProtocol);
      return 0;
    }

  /* Check endpoints */
  if (dev->config[0].interface[0].altsetting[0].bNumEndpoints
      != 2)
    {
      printf ("    this is not a MA-1015 (bNumEndpoints = %d)\n",
	      dev->config[0].interface[0].altsetting[0].bNumEndpoints);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress != 0x81)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize != 0x40)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval != 0x00))
    {
      printf ("    this is not a MA-1015 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[1].bEndpointAddress != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].wMaxPacketSize != 0x08)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].bInterval != 0x00))
    {
      printf ("    this is not a MA-1015 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bInterval);
      return 0;
    }
  
  /* Now we read register 0 to find out if this is really a MA-1015 */
  handle = usb_open (dev);
  if (handle == 0)
    {
      printf ("    Couldn't open device. Unload the scanner driver and run as root.\n");
      return 0;
    }
  result = usb_claim_interface (handle, 0);
  if (handle < 0)
    {
      printf ("    Couldn't claim interface. Unload the scanner driver and run as root.\n");
      usb_close (handle);
      return 0;
    }

  memset (req, 0, 8);
  req[0] = 33; req[1] = 0x00;
  result = usb_bulk_write (handle, 0x02, req, 8, 1000);
  if (result <= 0)
    {
      printf ("    this is not a MA-1015 (Error during bulk write)\n");
      usb_close (handle);
      return 0;
    }
  result = usb_bulk_read (handle, 0x81, &res, 1, 1000);
  if (result <= 0)
    {
      printf ("    this is not a MA-1015 (Error during bulk read)\n");
      usb_close (handle);
      return 0;
    }
  if (res != 0xa5)
    {
      printf ("    this is not a MA-1015 (got 0x%x, expected 0xa5)\n",
	      res);
      usb_close (handle);
      return 0;
    }

  usb_close (handle);
  printf ("  FOUND: Chipset is Mustek MA-1015\n");
  return 1;
}

/* Check for Mustek MA-1509 */
int
check_ma1509 (struct usb_device *dev)
{
  char req[8];
  unsigned char inquiry[0x80];
  usb_dev_handle *handle;
  int result;

  printf ("    checking for MA-1509 ...\n");

  /* Check device descriptor */
  if (dev->descriptor.bDeviceClass != USB_CLASS_VENDOR_SPEC)
    {
      printf ("    this is not a MA-1509 (bDeviceClass = %d)\n",
	      dev->descriptor.bDeviceClass);
      return 0;
    }
  if (dev->descriptor.bcdUSB != 0x110)
    {
      printf ("    this is not a MA-1509 (bcdUSB = 0x%x)\n",
	      dev->descriptor.bcdUSB);
      return 0;
    }
  if (dev->descriptor.bDeviceSubClass != 0xff)
    {
      printf ("    this is not a MA-1509 (bDeviceSubClass = 0x%x)\n",
	      dev->descriptor.bDeviceSubClass);
      return 0;
    }
  if (dev->descriptor.bDeviceProtocol != 0xff)
    {
      printf ("    this is not a MA-1509 (bDeviceProtocol = 0x%x)\n",
	      dev->descriptor.bDeviceProtocol);
      return 0;
    }

  /* Check endpoints */
  if (dev->config[0].interface[0].altsetting[0].bNumEndpoints
      != 3)
    {
      printf ("    this is not a MA-1509 (bNumEndpoints = %d)\n",
	      dev->config[0].interface[0].altsetting[0].bNumEndpoints);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress != 0x81)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize != 0x40)
      || (dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval != 0xff))
    {
      printf ("    this is not a MA-1509 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[0].bInterval);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[1].bEndpointAddress != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].bmAttributes != 0x02)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].wMaxPacketSize != 0x08)
      || (dev->config[0].interface[0].altsetting[0].endpoint[1].bInterval != 0x00))
    {
      printf ("    this is not a MA-1509 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[1].bInterval);
      return 0;
    }
  if ((dev->config[0].interface[0].altsetting[0].endpoint[2].bEndpointAddress != 0x83)
      || (dev->config[0].interface[0].altsetting[0].endpoint[2].bmAttributes != 0x03)
      || (dev->config[0].interface[0].altsetting[0].endpoint[2].wMaxPacketSize != 0x08)
      || (dev->config[0].interface[0].altsetting[0].endpoint[2].bInterval != 0x10))
    {
      printf ("    this is not a MA-1509 (bEndpointAddress = 0x%x, bmAttributes = 0x%x, "
	      "wMaxPacketSize = 0x%x, bInterval = 0x%x)\n",
	      dev->config[0].interface[0].altsetting[0].endpoint[2].bEndpointAddress,
	      dev->config[0].interface[0].altsetting[0].endpoint[2].bmAttributes,
	      dev->config[0].interface[0].altsetting[0].endpoint[2].wMaxPacketSize,
	      dev->config[0].interface[0].altsetting[0].endpoint[2].bInterval);
      return 0;
    }

  /* This is a SCSI-ober-USB chip, we'll read the inquiry */
  handle = usb_open (dev);
  if (handle == 0)
    {
      printf ("    Couldn't open device. Unload the scanner driver and run as root.\n");
      return 0;
    }
  result = usb_claim_interface (handle, 0);
  if (handle < 0)
    {
      printf ("    Couldn't claim interface. Unload the scanner driver and run as root.\n");
      usb_close (handle);
      return 0;
    }

  memset (req, 0, 8);
  req[0] = 0x12;
  req[1] = 1;
  req[6] = 0x60;

  result = usb_bulk_write (handle, 0x02, req, 8, 1000);
  if (result <= 0)
    {
      printf ("    this is not a MA-1509 (Error during bulk write)\n");
      usb_close (handle);
      return 0;
    }
  memset (inquiry, 0, 0x80);
  result = usb_bulk_read (handle, 0x81, inquiry, 0x40, 1000);
  if (result != 0x40)
    {
      printf ("    this is not a MA-1509 (Error during bulk read 1: %d)\n",
	      result);
      usb_close (handle);
      return 0;
    }
  result = usb_bulk_read (handle, 0x81, inquiry + 0x40, 0x40, 1000);
  if (result != 0x20)
    {
      printf ("    this is not a MA-1509 (Error during bulk read 2: %d)\n",
	      result);
      usb_close (handle);
      return 0;
    }
  if (inquiry[0] & 0x1f != 0x06)
    {
      printf ("    this is not a MA-1509 (inquiry [0] = %d)\n", inquiry[0]);
      usb_close (handle);
      return 0;
    }
  if (strncmp (inquiry + 8, "SCANNER ", 8) != 0)
    {
      inquiry [16] = 0;
      printf ("    this is not a MA-1509 (vendor=%s)\n", inquiry + 8);
      usb_close (handle);
      return 0;
    }
  inquiry [36] = 0;
  printf ("    MA-1509 version %s\n", inquiry + 32);
  usb_close (handle);
  printf ("  FOUND: Chipset is Mustek MA-1509\n");
  return 1;
}

int main ()
{
  struct usb_bus *bus;
  struct usb_device *dev;
  int found = 0;

  printf("Copyright (C) 2003 Henning Meier-Geinitz <henning@meier-geinitz.de>.\n");
  printf("This program comes with ABSOLUTELY NO WARRANTY.\n");
  printf("This is free software, and you are welcome to redistribute it ");
  printf("under certain\nconditions. See COPYING file for details.\n");
  printf("check-usb-scanner build %d\n\n", BUILD);

  usb_init ();
#ifdef DEBUG
  usb_set_debug (255);
#endif  
  usb_find_busses ();
  usb_find_devices ();

  /* Check all USB busses and devices */
  for (bus = usb_get_busses (); bus; bus = bus->next)
    {
      for (dev = bus->devices; dev; dev = dev->next)
	{
	  char *manufacturer, *product;

	  /* Print ids and vendor/product strings if available */
	  printf ("Checking device 0x%04x/0x%04x",
		  dev->descriptor.idVendor, dev->descriptor.idProduct);
	  if (dev->descriptor.iManufacturer 
	      && (manufacturer = get_string_descriptor (dev, dev->descriptor.iManufacturer)) != 0)
	    printf (" [%s]", manufacturer);
	  if (dev->descriptor.iProduct
	      && (product = get_string_descriptor (dev, dev->descriptor.iProduct)) != 0)
	    printf (" [%s]", product);
	  printf ("\n");
	  /* Happens sometimes with BSD */
	  if (!dev->config)
	    {
	      printf ("  IGNORED (device not configured)\n");
	      continue;
	    }
	  /* We are not interested in hubs */
	  if (dev->descriptor.idVendor == 0 || dev->descriptor.idProduct == 0)
	    {
	      printf ("  IGNORED (root hub)\n");
	      continue;
	    }
	  /* This is a heuristic for scanners */
	  switch (dev->descriptor.bDeviceClass)
	    {
	    case USB_CLASS_VENDOR_SPEC:
	      break;
	    case USB_CLASS_PER_INTERFACE:
	      {
		int ok = 0;
		int interface;

		for (interface = 0; interface < dev->config[0].bNumInterfaces; interface ++)
		  {
		    switch (dev->config[0].interface[interface].altsetting[0].bInterfaceClass)
		      {
		      case USB_CLASS_VENDOR_SPEC:
		      case USB_CLASS_PER_INTERFACE:
		      case 16:                /* data? */
			ok = 1;
			break;
		      default:
		      }
		  }
		if (!ok)
		  {
		    printf ("  IGNORED (bDeviceClass = %d, no suitable interfaces found)\n",
			    dev->descriptor.bDeviceClass);
		    continue;
		  }
		}
	      break;
	    default:
	      printf ("  IGNORED (bDeviceClass = %d)\n", dev->descriptor.bDeviceClass);
	      continue;
	    }

	  /* Now we check for specific chipsets */

	  found = check_gt6801 (dev);
	  if (found)
	    continue;

	  found = check_gt6816 (dev);
	  if (found)
	    continue;

	  found = check_ma1017 (dev);
	  if (found)
	    continue;

	  found = check_ma1015 (dev);
	  if (found)
	    continue;

	  found = check_ma1509 (dev);
	  if (found)
	    continue;

	  /* Check for more chipsets here */

	  printf ("  UNKNOWN chipset type\n");
	}
    }
  exit (0);
}
